# NOANYIO - since anyio.Cancelled does not exist
import trio


async def foo():
    ...


# except cancelled/baseexception are also critical
async def foo4():
    try:
        ...
    except trio.Cancelled:
        await foo()  # error: 8, Statement("trio.Cancelled", lineno-1)
    except:
        await foo()  # safe, since after trio.Cancelled

    try:
        ...
    except trio.Cancelled:
        await foo()  # error: 8, Statement("trio.Cancelled", lineno-1)
    except BaseException:
        await foo()  # safe, since after trio.Cancelled


async def foo5():
    try:
        ...
    except trio.Cancelled:
        with trio.CancelScope(deadline=30, shield=True):
            await foo()  # safe
    except BaseException:
        await foo()  # safe, since after trio.Cancelled



"""Contains Visitor103_104.

Never have an except Cancelled or except BaseException block with a code path that
doesn't re-raise the error.
103 is given when an except block exits without any raises, while 104 is given when
an improper raise, or other flow control, is encountered.
"""


from __future__ import annotations

import ast
from typing import Any

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import critical_except, error_class, iter_guaranteed_once

_trio103_common_msg = "{} block with a code path that doesn't re-raise the error."
_suggestion = " Consider adding an `except {}: raise` before this exception handler."
_suggestion_dict: dict[tuple[str, ...], str] = {
    ("anyio",): "anyio.get_cancelled_exc_class()",
    ("trio",): "trio.Cancelled",
}
_suggestion_dict[("anyio", "trio")] = "[" + "|".join(_suggestion_dict.values()) + "]"


@error_class
class Visitor103_104(Flake8TrioVisitor):
    error_codes = {
        "TRIO103": _trio103_common_msg,
        "TRIO104": "Cancelled (and therefore BaseException) must be re-raised.",
    }
    for poss_library in _suggestion_dict:
        error_codes[
            f"TRIO103_{'_'.join(poss_library)}"
        ] = _trio103_common_msg + _suggestion.format(_suggestion_dict[poss_library])

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.except_name: str | None = ""
        self.unraised: bool = False
        self.unraised_break: bool = False
        self.unraised_continue: bool = False
        self.loop_depth = 0

        self.cancelled_caught: set[str] = set()

    # If an `except` is bare, catches `BaseException`, or `trio.Cancelled`
    # set self.unraised, and if it's still set after visiting child nodes
    # then there might be a code path that doesn't re-raise.
    def visit_ExceptHandler(self, node: ast.ExceptHandler):
        marker = critical_except(node)

        if marker is None:
            return

        # If previous excepts have handled trio.Cancelled, don't do anything - namely
        # don't set self.unraised (so 104 isn't triggered) nor check for 103.
        if marker.name == "trio.Cancelled":
            error_code = "TRIO103"
            self.cancelled_caught.add("trio")
        elif marker.name in (
            "anyio.get_cancelled_exc_class()",
            "get_cancelled_exc_class()",
        ):
            error_code = "TRIO103"
            self.cancelled_caught.add("anyio")
        else:
            if self.cancelled_caught:
                return
            if len(self.library) < 2:
                error_code = f"TRIO103_{self.library_str}"
            else:
                error_code = f"TRIO103_{'_'.join(sorted(self.library))}"
            self.cancelled_caught.update("trio", "anyio")

        # Don't save the state of cancelled_caught, that's handled in Try and would
        # reset it between each except
        # Don't need to reset the values of unraised_[break|continue] since that's handled
        # by visit_For, but need to save the state of them to not mess up loops we're
        # nested inside
        self.save_state(
            node,
            "except_name",
            "unraised",
            "unraised_break",
            "unraised_continue",
            "loop_depth",
        )

        # save name from `as <except_name>`
        self.except_name = node.name

        self.loop_depth = 0
        self.unraised = True

        # Visit child nodes manually since we want to do logic afterwards.
        # Will unset self.unraised if all code paths `raise`
        self.generic_visit(node)

        if self.unraised:
            self.error(
                marker,
                marker.name,
                error_code=error_code,
            )

    def visit_Raise(self, node: ast.Raise):
        # if there's an unraised critical exception, the raise isn't bare,
        # and the name doesn't match, signal a problem.
        if (
            self.unraised
            and node.exc is not None
            and not (isinstance(node.exc, ast.Name) and node.exc.id == self.except_name)
        ):
            self.error(node, error_code="TRIO104")

        # treat it as safe regardless, to avoid unnecessary error messages.
        self.unraised = False

    def visit_Return(self, node: ast.Return | ast.Yield):
        if self.unraised:
            # Error: must re-raise
            self.error(node, error_code="TRIO104")

    visit_Yield = visit_Return

    # Treat Try's as fully covering only if `finally` always raises.
    def visit_Try(self, node: ast.Try):
        self.save_state(node, "cancelled_caught", copy=True)
        self.cancelled_caught = set()

        if not self.unraised:
            return

        # in theory it's okay if the try and all excepts re-raise,
        # and there is a bare except
        # but is a pain to parse and would require a special case for bare raises in
        # nested excepts.
        for n in (*node.body, *node.handlers, *node.orelse):
            self.visit(n)
            # re-set unraised to warn about returns in each block
            self.unraised = True

        # but it's fine if we raise in finally
        self.visit_nodes(node.finalbody)

    # Treat if's as fully covering if both `if` and `else` raise.
    # `elif` is parsed by the ast as a new if statement inside the else.
    def visit_If(self, node: ast.If):
        if not self.unraised:
            return

        body_raised = False
        self.visit_nodes(node.body)

        # does body always raise correctly
        body_raised = not self.unraised

        self.unraised = True
        self.visit_nodes(node.orelse)

        # if body didn't raise, or it's unraised after else, set unraise
        self.unraised = not body_raised or self.unraised

    # A loop is guaranteed to raise if:
    # condition always raises, or
    #   else always raises, and
    #   always raise before break
    # or body always raises (before break) and is guaranteed to run at least once
    def visit_For(self, node: ast.For | ast.While):
        if not self.unraised:
            return

        # the following block is duplicated in Visitor91X
        infinite_loop = False
        if isinstance(node, ast.While):
            try:
                infinite_loop = body_guaranteed_once = bool(ast.literal_eval(node.test))
            except Exception:  # noqa: PIE786
                body_guaranteed_once = False
            self.visit_nodes(node.test)
        else:
            self.visit_nodes(node.target)
            self.visit_nodes(node.iter)
            body_guaranteed_once = iter_guaranteed_once(node.iter)

        self.save_state(node, "unraised_break", "unraised_continue")
        self.unraised_break = False
        self.unraised_continue = False

        self.loop_depth += 1
        self.visit_nodes(node.body)
        self.loop_depth -= 1

        # if body is not guaranteed to run, or can continue at unraised, reset
        if not (infinite_loop or (body_guaranteed_once and not self.unraised_continue)):
            self.unraised = True
        self.visit_nodes(node.orelse)

        # if we might break at an unraised point, set unraised
        self.unraised |= self.unraised_break

    visit_While = visit_For

    def visit_Break(self, node: ast.Break):
        if self.unraised and self.loop_depth == 0:
            self.error(node, error_code="TRIO104")
        self.unraised_break |= self.unraised

    def visit_Continue(self, node: ast.Continue):
        if self.unraised and self.loop_depth == 0:
            self.error(node, error_code="TRIO104")
        self.unraised_continue |= self.unraised



# ARG --enable-visitor-codes-regex=(TRIO230)|(TRIO231)
import io
import os

import trio


async def foo():
    open("")  # TRIO230: 4, 'open', "trio"
    io.open_code("")  # TRIO230: 4, 'io.open_code', "trio"
    os.fdopen(0)  # TRIO231: 4, 'os.fdopen', "trio"

    # sync call is awaited, so it can't be sync
    await open("")

    # wrapped calls are always okay
    await trio.wrap_file(open(""))
    await trio.wrap_file(os.fdopen(0))

    # with uses the same code & logic
    with os.fdopen(0):  # TRIO231: 9, 'os.fdopen', "trio"
        ...
    with open(""):  # TRIO230: 9, 'open', "trio"
        ...
    with open("") as f:  # TRIO230: 9, 'open', "trio"
        ...
    with foo(), open(""):  # TRIO230: 16, 'open', "trio"
        ...
    async with open(""):  # TRIO230: 15, 'open', "trio"
        ...
    async with trio.wrap_file(open("")):
        ...

    # test io.open
    # pyupgrade removes the unnecessary `io.`
    # https://github.com/asottile/pyupgrade#open-alias
    # and afaict neither respects fmt:off nor #noqa - so I don't know how to test it
    open("")  # TRIO230: 4, 'open', "trio"


def foo_sync():
    open("")



"""Utility visitors for tracking shared state and modifying the tree."""

from __future__ import annotations

import ast

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import utility_visitor


@utility_visitor
class VisitorTypeTracker(Flake8TrioVisitor):
    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef | ast.FunctionDef | ast.Lambda
    ):
        def or_none(node: ast.AST | None):
            if not isinstance(node, ast.BinOp) or not isinstance(node.op, ast.BitOr):
                return None
            if isinstance(node.left, ast.Constant) and node.left.value is None:
                return node.right
            if isinstance(node.right, ast.Constant) and node.right.value is None:
                return node.left
            return None

        self.save_state(node, "variables", copy=True)

        args = node.args
        for arg in *args.args, *args.posonlyargs, *args.kwonlyargs:
            annotation = arg.annotation

            if (
                isinstance(annotation, ast.Subscript)
                and isinstance(annotation.value, ast.Name)
                and annotation.value.id == "Optional"
            ):
                annotation = annotation.slice
            elif res := or_none(annotation):
                annotation = res

            if isinstance(annotation, (ast.Name, ast.Attribute, ast.Constant)):
                annotation_type = ast.unparse(annotation)
            else:
                annotation_type = "Any"

            self.variables[arg.arg] = annotation_type

    visit_FunctionDef = visit_AsyncFunctionDef
    visit_Lambda = visit_AsyncFunctionDef

    def visit_AnnAssign(self, node: ast.AnnAssign):
        if not isinstance(node.target, ast.Name):
            return
        target = node.target.id
        typename = ast.unparse(node.annotation)
        self.variables[target] = typename

    def visit_Assign(self, node: ast.Assign):
        if len(node.targets) != 1 or not isinstance(node.targets[0], ast.Name):
            return

        # `f = open(...)`
        if isinstance(node.value, ast.Call) and (
            vartype := self.typed_calls.get(ast.unparse(node.value.func), None)
        ):
            self.variables[node.targets[0].id] = vartype

        # f = ff (and ff is a variable with known type)
        elif isinstance(node.value, ast.Name) and (
            value := self.variables.get(node.value.id, None)
        ):
            self.variables[node.targets[0].id] = value

    def visit_With(self, node: ast.With):
        if len(node.items) != 1:
            return
        item = node.items[0]
        if (
            isinstance(item.context_expr, ast.Call)
            and isinstance(item.context_expr.func, ast.Name)
            and isinstance(item.optional_vars, ast.Name)
            and (vartype := self.typed_calls.get(item.context_expr.func.id, None))
        ):
            self.variables[item.optional_vars.id] = vartype


@utility_visitor
class VisitorAwaitModifier(Flake8TrioVisitor):
    def visit_Await(self, node: ast.Await):
        if isinstance(node.value, ast.Call):
            # add attribute to indicate it's awaited
            setattr(node.value, "awaited", True)  # noqa: B010


@utility_visitor
class VisitorLibraryHandler(Flake8TrioVisitor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # check whether library we're working towards has been explicitly
        # specified with --anyio, otherwise assume Trio - but we update if we
        # see imports
        if self.options.anyio:
            self.add_library("anyio")

    def visit_Import(self, node: ast.Import):
        for alias in node.names:
            name = alias.name
            if name in ("trio", "anyio") and alias.asname is None:
                self.add_library(name)



import urllib3


async def foo_call():
    client = httpx.Client()
    client.close()  # TRIO212: 4, "close", "client"
    client.delete()  # TRIO212: 4, "delete", "client"
    client.get()  # TRIO212: 4, "get", "client"
    client.head()  # TRIO212: 4, "head", "client"
    client.options()  # TRIO212: 4, "options", "client"
    client.patch()  # TRIO212: 4, "patch", "client"
    client.post()  # TRIO212: 4, "post", "client"
    client.put()  # TRIO212: 4, "put", "client"
    client.request()  # TRIO212: 4, "request", "client"
    client.send()  # TRIO212: 4, "send", "client"
    client.stream()  # TRIO212: 4, "stream", "client"

    client.anything()
    client.build_request()
    client.is_closed


async def foo_ann_par(client: httpx.Client):
    client.send(...)  # TRIO212: 4, "send", "client"


async def foo_ann_par_leftNone(client: None | httpx.Client):
    client.send(...)  # TRIO212: 4, "send", "client"


async def foo_ann_par_rightNone(client: httpx.Client | None):
    client.send(...)  # TRIO212: 4, "send", "client"


async def foo_ann_par_optional(client: Optional[httpx.Client]):
    client.send(...)  # TRIO212: 4, "send", "client"


async def foo_ann_var():
    client: httpx.Client = ...
    client.send()  # TRIO212: 4, "send", "client"


glob_client = httpx.Client()


async def foo_global_client():
    glob_client.send()  # TRIO212: 4, "send", "glob_client"


def sync_global_client():
    glob_client.send()


glob_client: int = 4


async def foo_global_client_overriden():
    glob_client.send()


# not implemented
from httpx import Client


async def foo_alias_import():
    client: Client = Client()
    client.send()


# urllib3 pools
async def foo_urllib3():
    c: urllib3.HTTPConnectionPool = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.HTTPSConnectionPool = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.PoolManager = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.ProxyManager = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.connectionpool.ConnectionPool = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.connectionpool.HTTPConnectionPool = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.connectionpool.HTTPSConnectionPool = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.poolmanager.PoolManager = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.poolmanager.ProxyManager = ...
    c.anything()  # TRIO212: 4, "anything", "c"
    c: urllib3.request.RequestMethods = ...
    c.anything()  # TRIO212: 4, "anything", "c"

    c: urllib3.anything = ...
    c.anything()



"""A highly opinionated flake8 plugin for Trio-related problems.

This can include anything from outright bugs, to pointless/dead code,
to likely performance issues, to minor points of idiom that might signal
a misunderstanding.

It may well be too noisy for anyone with different opinions, that's OK.

Pairs well with flake8-async and flake8-bugbear.
"""

from __future__ import annotations

import ast
import keyword
import re
import tokenize
from argparse import ArgumentTypeError, Namespace
from typing import TYPE_CHECKING

from .runner import Flake8TrioRunner
from .visitors import default_disabled_error_codes

if TYPE_CHECKING:
    from collections.abc import Iterable
    from os import PathLike

    from flake8.options.manager import OptionManager

    from .base import Error


# CalVer: YY.month.patch, e.g. first release of July 2022 == "22.7.1"
__version__ = "23.2.3"


class Plugin:
    name = __name__
    version = __version__
    options: Namespace = Namespace()

    def __init__(self, tree: ast.AST):
        self._tree = tree

    @classmethod
    def from_filename(cls, filename: str | PathLike[str]) -> Plugin:  # pragma: no cover
        # only used with --runslow
        with tokenize.open(filename) as f:
            source = f.read()
        return cls(ast.parse(source))

    def run(self) -> Iterable[Error]:
        yield from Flake8TrioRunner.run(self._tree, self.options)

    @staticmethod
    def add_options(option_manager: OptionManager):
        # Disable TRIO9xx calls
        option_manager.extend_default_ignore(default_disabled_error_codes)

        option_manager.add_option(
            "--no-checkpoint-warning-decorators",
            default="asynccontextmanager",
            parse_from_config=True,
            required=False,
            comma_separated_list=True,
            help=(
                "Comma-separated list of decorators to disable TRIO910 & TRIO911 "
                "checkpoint warnings for. "
                "Decorators can be dotted or not, as well as support * as a wildcard. "
                "For example, ``--no-checkpoint-warning-decorators=app.route,"
                "mydecorator,mypackage.mydecorators.*``"
            ),
        )
        option_manager.add_option(
            "--startable-in-context-manager",
            type=parse_trio114_identifiers,
            default="",
            parse_from_config=True,
            required=False,
            help=(
                "Comma-separated list of method calls to additionally enable TRIO113 "
                "warnings for. Will also check for the pattern inside function calls. "
                "Methods must be valid identifiers as per `str.isidientifier()` and "
                "not reserved keywords. "
                "For example, ``--startable-in-context-manager=worker_serve,"
                "myfunction``"
            ),
        )
        option_manager.add_option(
            "--trio200-blocking-calls",
            type=parse_trio200_dict,
            default={},
            parse_from_config=True,
            required=False,
            help=(
                "Comma-separated list of key->value pairs, where key is a [dotted] "
                "function that if found inside an async function will raise TRIO200, "
                "suggesting it be replaced with {value}"
            ),
        )
        option_manager.add_option(
            "--enable-visitor-codes-regex",
            type=re.compile,
            default=".*",
            parse_from_config=True,
            required=False,
            help=(
                "Regex string of visitors to enable. Can be used to disable broken "
                "visitors, or instead of --select/--disable to select error codes "
                "in a way that is more performant. If a visitor raises multiple codes "
                "it will not be disabled unless all codes are disabled, but it will "
                "not report codes matching this regex."
            ),
        )
        option_manager.add_option(
            "--anyio",
            # action=store_true + parse_from_config does seem to work here, despite
            # https://github.com/PyCQA/flake8/issues/1770
            action="store_true",
            parse_from_config=True,
            required=False,
            default=False,
            help=(
                "Change the default library to be anyio instead of trio."
                " If trio is imported it will assume both are available and print"
                " suggestions with [anyio|trio]."
            ),
        )

    @staticmethod
    def parse_options(options: Namespace):
        Plugin.options = options


# flake8 ignores type parameters if using comma_separated_list
# so we need to reimplement that ourselves if we want to use "type"
# to check values
def parse_trio114_identifiers(raw_value: str) -> list[str]:
    values = [s.strip() for s in raw_value.split(",") if s.strip()]
    for value in values:
        if keyword.iskeyword(value) or not value.isidentifier():
            raise ArgumentTypeError(f"{value!r} is not a valid method identifier")
    return values


def parse_trio200_dict(raw_value: str) -> dict[str, str]:
    res: dict[str, str] = {}
    splitter = "->"  # avoid ":" because it's part of .ini file syntax
    values = [s.strip() for s in raw_value.split(",") if s.strip()]

    for value in values:
        split_values = list(map(str.strip, value.split(splitter)))
        if len(split_values) != 2:
            # argparse will eat this error message and spit out it's own
            # if we raise it as ValueError
            raise ArgumentTypeError(
                f"Invalid number ({len(split_values)-1}) of splitter "
                f"tokens {splitter!r} in {value!r}",
            )
        res[split_values[0]] = split_values[1]
    return res



"""Helper functions used in several visitor classes.

Also contains the decorator definitions used to register error classes.
"""

from __future__ import annotations

import ast
from fnmatch import fnmatch
from typing import TYPE_CHECKING, TypeVar

from ..base import Statement
from . import ERROR_CLASSES, default_disabled_error_codes, utility_visitors

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .flake8triovisitor import Flake8TrioVisitor, HasLineCol

    T = TypeVar("T", bound=Flake8TrioVisitor)


def error_class(error_class: type[T]) -> type[T]:
    ERROR_CLASSES.add(error_class)
    return error_class


def disabled_by_default(error_class: type[T]) -> type[T]:
    default_disabled_error_codes.extend(error_class.error_codes)
    return error_class


def utility_visitor(c: type[T]) -> type[T]:
    c.error_codes = {"noerror": "ignore"}
    utility_visitors.add(c)
    return c


# ignores module and only checks the unqualified name of the decorator
# used in 101 and 910/911
def has_decorator(decorator_list: list[ast.expr], *names: str):
    return any(
        (isinstance(dec, ast.Name) and dec.id in names)
        or (isinstance(dec, ast.Attribute) and dec.attr in names)
        for dec in decorator_list
    )


# matches the fully qualified name against fnmatch pattern
# used to match decorators and methods to user-supplied patterns
# used in 910/911 and 200
def fnmatch_qualified_name(name_list: list[ast.expr], *patterns: str) -> str | None:
    for name in name_list:
        if isinstance(name, ast.Call):
            name = name.func
        qualified_name = ast.unparse(name)

        for pattern in patterns:
            # strip leading "@"s for when we're working with decorators
            if fnmatch(qualified_name, pattern.lstrip("@")):
                return pattern
    return None


# used in 103/104 and 910/911
def iter_guaranteed_once(iterable: ast.expr) -> bool:
    # static container with an "elts" attribute
    if hasattr(iterable, "elts"):
        elts: Iterable[ast.expr] = iterable.elts  # type: ignore
        for elt in elts:
            assert isinstance(elt, ast.expr)
            # recurse starred expression
            if isinstance(elt, ast.Starred):
                if iter_guaranteed_once(elt.value):
                    return True
            else:
                return True
        return False

    if isinstance(iterable, ast.Constant):
        return hasattr(iterable.value, "__len__") and len(iterable.value) > 0

    if isinstance(iterable, ast.Dict):
        for key, val in zip(iterable.keys, iterable.values):
            # {**{...}, **{<...>}} is parsed as {None: {...}, None: {<...>}}
            if key is None and isinstance(val, ast.Dict):
                if iter_guaranteed_once(val):
                    return True
            else:
                return True
    # check for range() with literal parameters
    if (
        isinstance(iterable, ast.Call)
        and isinstance(iterable.func, ast.Name)
        and iterable.func.id == "range"
    ):
        try:
            return len(range(*[ast.literal_eval(a) for a in iterable.args])) > 0
        except Exception:  # noqa: PIE786
            return False
    return False


# used in 102, 103 and 104
def critical_except(node: ast.ExceptHandler) -> Statement | None:
    def has_exception(node: ast.expr) -> str | None:
        name = ast.unparse(node)
        if name in (
            "BaseException",
            "trio.Cancelled",
            "anyio.get_cancelled_exc_class()",
            "get_cancelled_exc_class()",
        ):
            return name
        return None

    name: str | None = None
    posnode: HasLineCol = node

    # bare except
    if node.type is None:
        name = "bare except"

    # several exceptions
    elif isinstance(node.type, ast.Tuple):
        for element in node.type.elts:
            name = has_exception(element)
            if name is not None:
                posnode = element
                break
    # single exception, either a Name or an Attribute
    else:
        name = has_exception(node.type)
        posnode = node.type

    if name is not None:
        return Statement(name, posnode.lineno, posnode.col_offset)

    return None


# used in 105 and 113
def is_nursery_call(node: ast.AST, name: str) -> bool:
    assert name in ("start", "start_soon")
    if isinstance(node, ast.Attribute):
        if isinstance(node.value, ast.Name):
            return node.attr == name and node.value.id.endswith("nursery")
        if isinstance(node.value, ast.Attribute):
            return node.attr == name and node.value.attr.endswith("nursery")
    return False


# used in 100, 101 and 102
cancel_scope_names = (
    "fail_after",
    "fail_at",
    "move_on_after",
    "move_on_at",
    "CancelScope",
)


# convenience function used in a lot of visitors
# should probably return a named tuple
def get_matching_call(
    node: ast.AST, *names: str, base: Iterable[str] = ("trio", "anyio")
) -> tuple[ast.Call, str, str] | None:
    if isinstance(base, str):
        base = (base,)
    if (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Attribute)
        and isinstance(node.func.value, ast.Name)
        and node.func.value.id in base
        and node.func.attr in names
    ):
        return node, node.func.attr, node.func.value.id
    return None



# ARG --enable-visitor-codes-regex=(TRIO103)|(TRIO104)

from typing import Any


def foo() -> Any:
    ...


# fmt: off
# TODO: Black 23.1.0 moves the long comments around a bit.

try:
    ...
except (SyntaxError, ValueError, BaseException):
    raise

try:
    ...
except (
    SyntaxError,
    ValueError,
    BaseException,  # TRIO103_trio: 4, "BaseException"
) as p:
    ...

try:
    ...
except (SyntaxError, ValueError):
    raise

try:
    ...
except BaseException as e:
    raise e

try:
    ...
except BaseException as e:
    raise  # acceptable - see https://peps.python.org/pep-0678/#example-usage

try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    ...

# if
try:
    ...
except BaseException as e:  # TRIO103_trio: 7, "BaseException"
    if True:
        raise e
    elif True:
        ...
    else:
        raise e

try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    if True:
        raise

try:
    ...
except BaseException:  # safe
    if True:
        raise
    elif True:
        raise
    else:
        raise

# loops
# raises inside the body are never guaranteed to run and are ignored
try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    while foo():
        raise

# raise inside else are guaranteed to run, unless there's a break
try:
    ...
except BaseException:
    while ...:
        ...
    else:
        raise

try:
    ...
except BaseException:
    for _ in "":
        ...
    else:
        raise

try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    while ...:
        if ...:
            break
        raise
    else:
        raise

try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    for _ in "":
        if ...:
            break
        raise
    else:
        raise

# ensure we don't ignore previous guaranteed raise (although that's unreachable code)
try:
    ...
except BaseException:
    raise
    for _ in "":
        if ...:
            break
        raise
    else:
        raise

# nested try
# in theory safe if the try, and all excepts raises - and there's a bare except.
# But is a very weird pattern that we don't handle.
try:
    ...
except BaseException as e:  # TRIO103_trio: 7, "BaseException"
    try:
        raise e
    except ValueError:
        raise e
    except:
        raise e  # TRIO104: 8

try:
    ...
except BaseException:  # safe
    try:
        ...
    finally:
        raise

# check that nested non-critical exceptions are ignored
try:
    ...
except BaseException:
    try:
        ...
    except ValueError:
        ...  # safe
    raise

# check that name isn't lost
try:
    ...
except BaseException as e:
    try:
        ...
    except BaseException as f:
        raise f
    raise e

# don't bypass raise by raising from nested except
try:
    ...
except BaseException as e:
    try:
        ...
    except ValueError as g:
        raise g  # TRIO104: 8
    except BaseException as h:
        raise h  # error? currently treated as safe
    raise e

# bare except, equivalent to `except baseException`
try:
    ...
except:  # TRIO103_trio: 0, "bare except"
    ...

try:
    ...
except:
    raise

# point to correct exception in multi-line handlers
my_super_mega_long_exception_so_it_gets_split = SyntaxError
try:
    ...
except (
    my_super_mega_long_exception_so_it_gets_split,
    SyntaxError,
    BaseException,  # TRIO103_trio: 4, "BaseException"
    ValueError,
    BaseException,  # no complaint on this line
):
    ...

# loop over non-empty static collection
try:
    ...
except BaseException as e:
    for i in [1, 2, 3]:
        raise

try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    for i in [1, 2, 3]:
        ...

try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    for i in [1, 2, 3]:
        if ...:
            continue
        raise

try:
    ...
except BaseException:
    while True:
        raise

try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    while True:
        if ...:
            break
        raise

try:
    ...
except BaseException:
    while True:
        if ...:
            continue
        raise

# Issue #106, false alarm on excepts after `Cancelled` has already been handled
try:
    ...
except BaseException:
    raise
except:  # now silent
    ...

# don't throw multiple 103's even if `Cancelled` wasn't properly handled.
try:
    ...
except BaseException:  # TRIO103_trio: 7, "BaseException"
    ...
except:  # now silent
    ...

# Check state management of whether cancelled has been handled across nested try's
try:
    try:
        ...
    except BaseException:
        raise
except BaseException:  # TRIO103_trio: 7, "BaseException"
    ...
except:
    try:
        ...
    except BaseException:  # TRIO103_trio: 11, "BaseException"
        ...



"""Tests for flake8-trio package metadata."""

from __future__ import annotations

import re
import unittest
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

import flake8_trio

from .test_flake8_trio import ERROR_CODES

if TYPE_CHECKING:
    from collections.abc import Iterable

root_path = Path(__file__).parent.parent


class Version(NamedTuple):
    year: int
    month: int
    patch: int

    @classmethod
    def from_string(cls, string):
        return cls(*map(int, string.split(".")))


def get_releases() -> Iterable[Version]:
    valid_pattern = re.compile(r"^## (\d\d\.\d?\d\.\d?\d)$")
    with open(root_path / "CHANGELOG.md", encoding="utf-8") as f:
        lines = f.readlines()
    for line in lines:
        version_match = valid_pattern.match(line)
        if version_match:
            yield Version.from_string(version_match.group(1))


def test_last_release_against_changelog():
    """Ensure we have the latest version covered in 'CHANGELOG.md'."""
    latest_release, *_ = get_releases()
    assert latest_release == Version.from_string(flake8_trio.__version__)


def test_version_increments_are_correct():
    versions = list(get_releases())
    for prev, current in zip(versions[1:], versions):
        assert prev < current  # remember that `versions` is newest-first
        msg = f"{current=} does not follow {prev=}"
        # for CalVer, we either increment the patch version by one, or
        # increment the time-based parts and set the patch version to one.
        if current.patch == 1:
            assert prev[:2] < current[:2], msg
        else:
            assert current == prev._replace(patch=prev.patch + 1), msg


IGNORED_CODES_REGEX = r"TRIO107|TRIO108|TRIO\d\d\d_.*"


class test_messages_documented(unittest.TestCase):
    def runTest(self):
        documented_errors: dict[str, set[str]] = {}
        for filename in (
            "CHANGELOG.md",
            "README.md",
        ):
            with open(root_path / filename, encoding="utf-8") as f:
                lines = f.readlines()
            documented_errors[filename] = set()
            for line in lines:
                for error_msg in re.findall(r"TRIO\d\d\d", line):
                    documented_errors[filename].add(error_msg)

        documented_errors["flake8_trio.py"] = set(ERROR_CODES)
        # check files for @error_class

        # get tested error codes from file names and from `INCLUDE` lines
        documented_errors["eval_files"] = set()
        p = Path(__file__).parent / "eval_files"
        for file_path in p.iterdir():
            if not file_path.is_file():
                continue

            if m := re.search(r"trio\d\d\d", str(file_path)):
                documented_errors["eval_files"].add(m.group().upper())

            with open(file_path) as file:
                for line in file:
                    if line.startswith("# ARG --enable-visitor-codes-regex"):
                        for m in re.findall(r"trio\d\d\d", line, re.IGNORECASE):
                            # pyright types m as `Any` (as it is in typeshed)
                            # mypy types it as Optional[Match[str]]
                            # but afaict it should be something like str|Tuple[str,...]
                            # depending on whether there's a group in the pattern or not.
                            # (or bytes, if both inputs are bytes)
                            assert isinstance(m, str)
                            documented_errors["eval_files"].add(m)
                        break

        for errset in documented_errors.values():
            errset.difference_update(
                [c for c in errset if re.fullmatch(IGNORED_CODES_REGEX, c)]
            )

        unique_errors: dict[str, set[str]] = {}
        missing_errors: dict[str, set[str]] = {}
        for key, codes in documented_errors.items():
            unique_errors[key] = codes.copy()
            missing_errors[key] = set()

            for other_key, other_codes in documented_errors.items():
                if key == other_key:
                    continue
                unique_errors[key].difference_update(other_codes)
                missing_errors[key].update(other_codes - codes)

        assert unique_errors == missing_errors



# ARG --no-checkpoint-warning-decorator=asynccontextmanager,other_context_manager


async def foo1():  # TRIO900: 0
    yield
    yield


@asynccontextmanager
async def foo2():
    yield


@asynccontextmanager
async def foo3():
    async def bar():  # TRIO900: 4
        yield

    yield


def foo4():
    yield


@pytest.fixture
async def async_fixtures_are_basically_context_managers():
    yield


@pytest.fixture(scope="function")  # args don't matter
async def async_fixtures_can_take_arguments():
    yield


# no-checkpoint-warning-decorator now ignored
@other_context_manager
async def foo5():  # TRIO900: 0
    yield



# NOTRIO
import anyio


# nurseries don't exist in anyio, so don't error on it.
async def foo():
    nursery = anyio.open_nursery()
    await nursery.start()
    await nursery.start_foo()

    nursery.start()  # should not be triggered with anyio
    None.start()
    nursery.start_soon()
    nursery.start_foo()



"""Visitor for TRIO118.

Don't assign the value of `anyio.get_cancelled_exc_class()` to a variable, since
that breaks linter checks and multi-backend programs.
"""


from __future__ import annotations

import ast
import re

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import error_class


@error_class
class Visitor118(Flake8TrioVisitor):
    error_codes = {
        "TRIO118": (
            "Don't assign the value of `anyio.get_cancelled_exc_class()` to a variable,"
            " since that breaks linter checks and multi-backend programs."
        )
    }

    def visit_Assign(self, node: ast.Assign | ast.AnnAssign):
        if node.value is None:
            return
        name = ast.unparse(node.value)
        if re.fullmatch(r"(anyio.)?get_cancelled_exc_class(\(\))?", name):
            self.error(node.value)

    visit_AnnAssign = visit_Assign

    # redundant check with TRIO106
    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module == "anyio":
            for alias in node.names:
                if alias.name == "get_cancelled_exc_class" and alias.asname is not None:
                    # alias doesn't have a lineno in 3.9
                    self.error(node)
                    return



# NOANYIO - don't run it with substitutions
import anyio
import trio

try:
    ...
except trio.Cancelled:  # TRIO103: 7, "trio.Cancelled"
    ...
except anyio.get_cancelled_exc_class():  # TRIO103: 7, "anyio.get_cancelled_exc_class()"
    ...
except:  # safe
    ...

try:
    ...
except anyio.get_cancelled_exc_class():  # TRIO103: 7, "anyio.get_cancelled_exc_class()"
    ...
except trio.Cancelled:  # TRIO103: 7, "trio.Cancelled"
    ...
except:  # safe
    ...

try:
    ...
except anyio.get_cancelled_exc_class():  # TRIO103: 7, "anyio.get_cancelled_exc_class()"
    ...
except:  # safe ?
    ...

try:
    ...
except trio.Cancelled:  # TRIO103: 7, "trio.Cancelled"
    ...
except:  # safe ?
    ...

# Check we get the proper suggestion when both are imported
try:
    ...
except BaseException:  # TRIO103_anyio_trio: 7, "BaseException"
    ...

try:
    ...
except:  # TRIO103_anyio_trio: 0, "bare except"
    ...



"""Check that all visitors are imported.

Checks that all flake8_trio/visitor*.py files are imported in flake8_trio/visitor/__init__
so their decorators are run.
"""

from __future__ import annotations

import re
from pathlib import Path


def test_all_visitors_imported():
    visitor_dir = Path(__file__).parent.parent / "flake8_trio" / "visitors"
    visitor_files = {
        f.stem for f in visitor_dir.iterdir() if f.stem.startswith("visitor")
    }
    visited_files = set()
    with open(visitor_dir / "__init__.py") as f:
        for line in f:
            if m := re.match(r"from \. import (?P<module>\w*)", line):
                visited_files.add(m.group("module"))
    assert visitor_files == visited_files



"""Contains Flake8TrioRunner.

The runner is what's run by the Plugin, and handles traversing
the AST and letting all registered ERROR_CLASSES do their visit'ing on them.
"""

from __future__ import annotations

import ast
import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .visitors import ERROR_CLASSES, utility_visitors

if TYPE_CHECKING:
    from argparse import Namespace
    from collections.abc import Iterable

    from .base import Error
    from .visitors.flake8triovisitor import Flake8TrioVisitor


@dataclass
class SharedState:
    options: Namespace
    problems: list[Error] = field(default_factory=list)
    library: tuple[str, ...] = field(default_factory=tuple)
    typed_calls: dict[str, str] = field(default_factory=dict)
    variables: dict[str, str] = field(default_factory=dict)


class Flake8TrioRunner(ast.NodeVisitor):
    def __init__(self, options: Namespace):
        super().__init__()
        self.state = SharedState(options)

        # utility visitors that need to run before the error-checking visitors
        self.utility_visitors = {v(self.state) for v in utility_visitors}

        self.visitors = {
            v(self.state) for v in ERROR_CLASSES if self.selected(v.error_codes)
        }

    def selected(self, error_codes: dict[str, str]) -> bool:
        return any(
            re.match(self.state.options.enable_visitor_codes_regex, code)
            for code in error_codes
        )

    @classmethod
    def run(cls, tree: ast.AST, options: Namespace) -> Iterable[Error]:
        runner = cls(options)
        runner.visit(tree)
        yield from runner.state.problems

    def visit(self, node: ast.AST):
        """Visit a node."""
        # tracks the subclasses that, from this node on, iterated through it's subfields
        # we need to remember it so we can restore it at the end of the function.
        novisit: set[Flake8TrioVisitor] = set()

        method = "visit_" + node.__class__.__name__

        for subclass in *self.utility_visitors, *self.visitors:
            # check if subclass has defined a visitor for this type
            class_method = getattr(subclass, method, None)
            if class_method is None:
                continue

            # call it
            class_method(node)

            # it will set `.novisit` if it has itself handled iterating through subfields
            # so we add it to our novisit set
            if subclass.novisit:
                novisit.add(subclass)

        # Remove all subclasses that iterated through subfields from our list of
        # visitors, so we don't visit them twice.
        self.visitors.difference_update(novisit)

        # iterate through subfields using NodeVisitor
        self.generic_visit(node)

        # reset the novisit flag for the classes in novisit
        for subclass in novisit:
            subclass.novisit = False

        # and add them back to our visitors
        self.visitors.update(novisit)

        # restore any outer state that was saved in the visitor method
        for subclass in *self.utility_visitors, *self.visitors:
            subclass.set_state(subclass.outer.pop(node, {}))



"""Various visitors/error classes that are too small to warrant getting their own file."""

from __future__ import annotations

import ast
from typing import Any, cast

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import (
    cancel_scope_names,
    disabled_by_default,
    error_class,
    get_matching_call,
    has_decorator,
    is_nursery_call,
)

# used in 100
checkpoint_node_types = (ast.Await, ast.AsyncFor, ast.AsyncWith)


@error_class
class Visitor100(Flake8TrioVisitor):
    error_codes = {
        "TRIO100": (
            "{0}.{1} context contains no checkpoints, remove the context or add"
            " `await {0}.lowlevel.checkpoint()`."
        ),
    }

    def visit_With(self, node: ast.With | ast.AsyncWith):
        for item in (i.context_expr for i in node.items):
            call = get_matching_call(item, *cancel_scope_names)
            if call and not any(
                isinstance(x, checkpoint_node_types) and x != node
                for x in ast.walk(node)
            ):
                # type checking has been done inside get_matching_call
                self.error(item, call[2], call[1])

    visit_AsyncWith = visit_With


@error_class
class Visitor101(Flake8TrioVisitor):
    error_codes = {
        "TRIO101": (
            "`yield` inside a nursery or cancel scope is only safe when implementing "
            "a context manager - otherwise, it breaks exception handling."
        ),
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._yield_is_error = False
        self._safe_decorator = False

    def visit_With(self, node: ast.With | ast.AsyncWith):
        self.save_state(node, "_yield_is_error", copy=True)
        for item in node.items:
            # if there's no safe decorator,
            # and it's not yet been determined that yield is error
            # and this withitem opens a cancelscope:
            # then yielding is unsafe
            if (
                not self._safe_decorator
                and not self._yield_is_error
                and get_matching_call(
                    item.context_expr, "open_nursery", *cancel_scope_names
                )
                is not None
            ):
                self._yield_is_error = True

    def visit_FunctionDef(self, node: ast.FunctionDef | ast.AsyncFunctionDef):
        self.save_state(node)
        self._yield_is_error = False
        self._safe_decorator = has_decorator(
            node.decorator_list, "contextmanager", "asynccontextmanager"
        )

    visit_AsyncWith = visit_With
    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_Yield(self, node: ast.Yield):
        if self._yield_is_error:
            self.error(node)


@error_class
class Visitor106(Flake8TrioVisitor):
    error_codes = {
        "TRIO106": "{0} must be imported with `import {0}` for the linter to work.",
    }

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module in ("trio", "anyio"):
            self.error(node, node.module)

    def visit_Import(self, node: ast.Import):
        for name in node.names:
            if name.name in ("trio", "anyio") and name.asname is not None:
                self.error(node, name.name)


@error_class
class Visitor109(Flake8TrioVisitor):
    error_codes = {
        "TRIO109": (
            "Async function definition with a `timeout` parameter - use "
            "`{}.[fail/move_on]_[after/at]` instead."
        ),
    }

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        # pending configuration or a more sophisticated check, ignore
        # all functions with a decorator
        if node.decorator_list:
            return

        args = node.args
        for arg in (*args.posonlyargs, *args.args, *args.kwonlyargs):
            if arg.arg == "timeout":
                self.error(arg, self.library_str)


@error_class
class Visitor110(Flake8TrioVisitor):
    error_codes = {
        "TRIO110": (
            "`while <condition>: await {0}.sleep()` should be replaced by "
            "a `{0}.Event`."
        ),
    }

    def visit_While(self, node: ast.While):
        if (
            len(node.body) == 1
            and isinstance(node.body[0], ast.Expr)
            and isinstance(node.body[0].value, ast.Await)
            and get_matching_call(node.body[0].value.value, "sleep", "sleep_until")
        ):
            self.error(node, self.library_str)


@error_class
class Visitor112(Flake8TrioVisitor):
    error_codes = {
        "TRIO112": (
            "Redundant nursery {}, consider replacing with directly awaiting "
            "the function call."
        ),
    }

    # if with has a withitem `trio.open_nursery() as <X>`,
    # and the body is only a single expression <X>.start[_soon](),
    # and does not pass <X> as a parameter to the expression
    def visit_With(self, node: ast.With | ast.AsyncWith):
        # body is single expression
        if len(node.body) != 1 or not isinstance(node.body[0], ast.Expr):
            return
        for item in node.items:
            # get variable name <X>
            if not isinstance(item.optional_vars, ast.Name):
                continue
            var_name = item.optional_vars.id

            # check for trio.open_nursery
            nursery = get_matching_call(item.context_expr, "open_nursery")

            # `isinstance(..., ast.Call)` is done in get_matching_call
            body_call = cast(ast.Call, node.body[0].value)

            if (
                nursery is not None
                and get_matching_call(body_call, "start", "start_soon", base=var_name)
                # check for presence of <X> as parameter
                and not any(
                    (isinstance(n, ast.Name) and n.id == var_name)
                    for n in self.walk(*body_call.args, *body_call.keywords)
                )
            ):
                self.error(item.context_expr, var_name)

    visit_AsyncWith = visit_With


# used in 113 and 900
def _get_identifier(node: ast.expr) -> str:
    if isinstance(node, ast.Name):
        return node.id
    if isinstance(node, ast.Attribute):
        return node.attr
    if isinstance(node, ast.Call):
        return _get_identifier(node.func)
    return ""


# used in 113 and 114
STARTABLE_CALLS = (
    "run_process",
    "serve_ssl_over_tcp",
    "serve_tcp",
    "serve_listeners",
    "serve",
)


@error_class
class Visitor113(Flake8TrioVisitor):
    error_codes = {
        "TRIO113": (
            "Dangerous `.start_soon()`, function might not be executed before"
            " `__aenter__` exits. Consider replacing with `.start()`."
        ),
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.async_function = False
        self.asynccontextmanager = False
        self.aenter = False

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        self.save_state(node, "aenter")

        self.aenter = node.name == "__aenter__" or any(
            _get_identifier(d) == "asynccontextmanager" for d in node.decorator_list
        )

    def visit_Yield(self, node: ast.Yield):
        self.aenter = False

    def visit_Call(self, node: ast.Call) -> None:
        def is_startable(n: ast.expr, *startable_list: str) -> bool:
            if isinstance(n, ast.Name):
                return n.id in startable_list
            if isinstance(n, ast.Attribute):
                return n.attr in startable_list
            if isinstance(n, ast.Call):
                return any(is_startable(nn, *startable_list) for nn in n.args)
            return False

        if (
            self.aenter
            and is_nursery_call(node.func, "start_soon")
            and len(node.args) > 0
            and is_startable(
                node.args[0],
                *STARTABLE_CALLS,
                *self.options.startable_in_context_manager,
            )
        ):
            self.error(node)


# Checks that all async functions with a "task_status" parameter have a match in
# --startable-in-context-manager. Will only match against the last part of the option
# name, so may miss cases where functions are named the same in different modules/classes
# and option names are specified including the module name.
@error_class
class Visitor114(Flake8TrioVisitor):
    error_codes = {
        "TRIO114": (
            "Startable function {} not in --startable-in-context-manager parameter "
            "list, please add it so TRIO113 can catch errors using it."
        ),
    }

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        if any(
            isinstance(n, ast.arg) and n.arg == "task_status"
            for n in self.walk(*node.args.args, *node.args.kwonlyargs)
        ) and not any(
            node.name == opt
            for opt in (*self.options.startable_in_context_manager, *STARTABLE_CALLS)
        ):
            self.error(node, node.name)


# Suggests replacing all `trio.sleep(0)` with the more suggestive
# `trio.lowlevel.checkpoint()`
@error_class
class Visitor115(Flake8TrioVisitor):
    error_codes = {
        "TRIO115": "Use `{0}.lowlevel.checkpoint()` instead of `{0}.sleep(0)`.",
    }

    def visit_Call(self, node: ast.Call):
        if (
            (m := get_matching_call(node, "sleep"))
            and len(node.args) == 1
            and isinstance(node.args[0], ast.Constant)
            and node.args[0].value == 0
        ):
            # m[2] is set to node.func.value.id
            self.error(node, m[2])


@error_class
class Visitor116(Flake8TrioVisitor):
    error_codes = {
        "TRIO116": (
            "{0}.sleep() with >24 hour interval should usually be "
            "`{0}.sleep_forever()`."
        ),
    }

    def visit_Call(self, node: ast.Call):
        if (m := get_matching_call(node, "sleep")) and len(node.args) == 1:
            arg = node.args[0]
            if (
                # `trio.sleep(math.inf)`
                (isinstance(arg, ast.Attribute) and arg.attr == "inf")
                # `trio.sleep(inf)`
                or (isinstance(arg, ast.Name) and arg.id == "inf")
                # `trio.sleep(float("inf"))`
                or (
                    isinstance(arg, ast.Call)
                    and isinstance(arg.func, ast.Name)
                    and arg.func.id == "float"
                    and len(arg.args)
                    and isinstance(arg.args[0], ast.Constant)
                    and arg.args[0].value == "inf"
                )
                # `trio.sleep(1e999)` (constant value inf)
                # `trio.sleep(86401)`
                # `trio.sleep(86400.1)`
                or (
                    isinstance(arg, ast.Constant)
                    and isinstance(arg.value, (int, float))
                    and arg.value > 86400
                )
            ):
                self.error(node, m[2])


DEPRECATED_ERRORS = ("MultiError", "NonBaseMultiError")


# anyio does not have MultiError, so this check is trio-only
@error_class
class Visitor117(Flake8TrioVisitor):
    error_codes = {
        "TRIO117": ("Reference to {}, prefer [exceptiongroup.]BaseExceptionGroup."),
    }

    # This should never actually happen given TRIO106
    def visit_Name(self, node: ast.Name):
        if node.id in DEPRECATED_ERRORS and "trio" in self.library:
            self.error(node, node.id)

    def visit_Attribute(self, node: ast.Attribute):
        if (n := ast.unparse(node)) in ("trio.MultiError", "trio.NonBaseMultiError"):
            self.error(node, n)


@error_class
@disabled_by_default
class Visitor900(Flake8TrioVisitor):
    error_codes = {
        "TRIO900": ("Async generator without `@asynccontextmanager` not allowed.")
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.unsafe_function: ast.AsyncFunctionDef | None = None

    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef | ast.FunctionDef | ast.Lambda
    ):
        self.save_state(node, "unsafe_function")
        if isinstance(node, ast.AsyncFunctionDef) and not any(
            _get_identifier(d) in ("asynccontextmanager", "fixture")
            for d in node.decorator_list
        ):
            self.unsafe_function = node

    def visit_Yield(self, node: ast.Yield):
        if self.unsafe_function is not None:
            self.error(self.unsafe_function)
            self.unsafe_function = None

    visit_FunctionDef = visit_AsyncFunctionDef
    visit_Lambda = visit_AsyncFunctionDef



# ARG --enable-visitor-codes-regex=TRIO103
# check that partly disabling a visitor works
from typing import Any


def foo() -> Any:
    ...


# nested try
# in theory safe if the try, and all excepts raises - and there's a bare except.
# But is a very weird pattern that we don't handle.
try:
    ...
except BaseException as e:  # TRIO103_trio: 7, "BaseException"
    try:
        raise e
    except ValueError:
        raise e
    except:
        raise e  # disabled TRIO104 error



"""TRIO105: library async function must be immediately awaited."""

from __future__ import annotations

import ast

from .flake8triovisitor import Flake8TrioVisitor
from .helpers import error_class, is_nursery_call

# used in 105
trio_async_funcs = (
    "aclose_forcefully",
    "open_file",
    "open_ssl_over_tcp_listeners",
    "open_ssl_over_tcp_stream",
    "open_tcp_listeners",
    "open_tcp_stream",
    "open_unix_socket",
    "run_process",
    "serve_listeners",
    "serve_ssl_over_tcp",
    "serve_tcp",
    "sleep",
    "sleep_forever",
    "sleep_until",
)


@error_class
class Visitor105(Flake8TrioVisitor):
    error_codes = {
        "TRIO105": "{1} async function {0} must be immediately awaited.",
    }

    def visit_Call(self, node: ast.Call):
        if "trio" not in self.library:
            return
        if (
            not getattr(node, "awaited", False)
            and isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and (
                (node.func.value.id == "trio" and node.func.attr in trio_async_funcs)
                or is_nursery_call(node.func, "start")
            )
        ):
            self.error(node, node.func.attr, node.func.value.id)
